from collections import UserString
from typing import Any, Iterator, TypedDict

from itemadapter.adapter import ItemAdapter
from itemadapter.utils import is_item
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.spiders import Spider
from scrapy.utils.misc import build_from_crawler, load_object

from scrapy_translate.processors import (
    extract_text_from_html,
    extract_text_from_list,
    inject_text_into_html,
    inject_text_into_list,
)
from scrapy_translate.providers import (
    CacheProvider,
    NullCacheProvider,
    TranslationProvider,
)


class TranslatedString(UserString):
    cache: bool

    def __init__(self, *args, cache: bool = False) -> None:
        super().__init__(*args)
        self.cache = cache


class TranslatedFieldMeta(TypedDict):
    html: bool
    cache: bool


class TranslatePipeline:
    _cache_provider: CacheProvider
    _translation_provider: TranslationProvider

    def __init__(
        self,
        *,
        cache_provider: CacheProvider,
        translation_provider: TranslationProvider,
    ):
        self._cache_provider = cache_provider
        self._translation_provider = translation_provider

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> "TranslatePipeline":
        if crawler.settings.getbool("TRANSLATION_DISABLED"):
            raise NotConfigured

        if not (translation_provider := crawler.settings.get("TRANSLATION_PROVIDER")):
            raise NotConfigured("TRANSLATION_PROVIDER must be set")
        cache_provider = crawler.settings.setdefault(
            "TRANSLATION_CACHE_PROVIDER",
            NullCacheProvider,
        )

        return cls(
            cache_provider=build_from_crawler(load_object(cache_provider), crawler),
            translation_provider=build_from_crawler(
                load_object(translation_provider),
                crawler,
            ),
        )

    async def process_item(self, item: Any, spider: Spider) -> Any:
        if not is_item(item):
            return item

        item_adapter = ItemAdapter(item)
        translated_strings = self._get_translated_strings(item_adapter)
        translations = await self._translate(translated_strings)
        self._set_translated_strings(item_adapter, translations)
        return item_adapter.item

    def _translated_fields_meta(
        self,
        item_adapter: ItemAdapter,
    ) -> Iterator[tuple[str, TranslatedFieldMeta]]:
        for field_name in item_adapter.field_names():
            field_meta = item_adapter.get_field_meta(field_name)
            if field_meta.get("translate"):
                yield (
                    field_name,
                    TranslatedFieldMeta(
                        html=field_meta.get("translate_html", False),
                        cache=field_meta.get("translate_cache", False),
                    ),
                )

    def _get_translated_strings(
        self,
        item_adapter: ItemAdapter,
    ) -> list[TranslatedString]:
        translated_strings: list[TranslatedString] = []
        for field_name, field_meta in self._translated_fields_meta(item_adapter):
            if field_meta["html"]:
                strings = extract_text_from_html(item_adapter[field_name])
            elif isinstance(item_adapter[field_name], list):
                strings = extract_text_from_list(item_adapter[field_name])
            else:
                strings = [item_adapter[field_name]]

            translated_strings.extend(
                TranslatedString(string, cache=field_meta.get("cache", False))
                for string in strings
                if string
            )

        return translated_strings

    def _set_translated_strings(
        self,
        item_adapter: ItemAdapter,
        translations: dict[str, str],
    ) -> None:
        for field_name, field_meta in self._translated_fields_meta(item_adapter):
            if field_meta.get("html"):
                item_adapter[field_name] = inject_text_into_html(
                    item_adapter[field_name],
                    [
                        translations.get(string, string)
                        for string in extract_text_from_html(item_adapter[field_name])
                    ],
                )
            elif isinstance(item_adapter[field_name], list):
                item_adapter[field_name] = inject_text_into_list(
                    item_adapter[field_name],
                    [
                        translations.get(string, string)
                        for string in extract_text_from_list(item_adapter[field_name])
                    ],
                )
            else:
                item_adapter[field_name] = translations.get(
                    item_adapter[field_name], item_adapter[field_name]
                )

    async def _translate(self, strings: list[TranslatedString]) -> dict[str, str]:
        cached_translations: dict[str, str] = {}
        from_cache = set(s.data for s in strings if s.cache)
        if from_cache:
            cached_translations = await self._cache_provider.get(from_cache)

        translations: dict[str, str] = {}
        to_translate = set(s.data for s in strings if s.data not in cached_translations)
        if to_translate:
            translations = await self._translation_provider.translate(to_translate)
            to_cache = {
                k: v
                for k, v in translations.items()
                if k in from_cache and k not in cached_translations
            }
            if to_cache:
                await self._cache_provider.set(to_cache)

        translations.update(cached_translations)
        return translations



import unittest

from scrapy_translate.processors import (
    extract_text_from_html,
    extract_text_from_list,
    inject_text_into_html,
    inject_text_into_list,
)


class TestProcessors(unittest.TestCase):
    def test_html_processor(self):
        html = '<p>Hello, <strong>World!</strong><br><img src="img.png"></p>'
        result = inject_text_into_html(html, extract_text_from_html(html))
        self.assertEqual(html, result)

    def test_list_processor(self):
        checks = ([], ["foo"], ["foo", "bar"], [["foo", "bar"], "baz"])
        for value in checks:
            with self.subTest(value=value):
                result = inject_text_into_list(value, extract_text_from_list(value))
                self.assertEqual(value, result)



import io
from html.parser import HTMLParser
from typing import Iterator, Optional


class HTMLTextExtractor(HTMLParser):
    _buffer: list[str]

    def __init__(self) -> None:
        super().__init__(convert_charrefs=True)

    def set_buffer(self, new_buffer: list[str]) -> None:
        self._buffer = new_buffer

    def handle_data(self, data: str) -> None:
        self._buffer.append(data)


class HTMLTextInjector(HTMLParser):
    _buffer: io.StringIO
    _text: Iterator[str]

    def __init__(self) -> None:
        super().__init__(convert_charrefs=True)

    def set_buffer(self, new_buffer: io.StringIO) -> None:
        self._buffer = new_buffer

    def set_text(self, text: list[str]) -> None:
        self._text = iter(text)

    def handle_starttag(
        self,
        tag: str,
        attrs: list[tuple[str, Optional[str]]],
    ) -> None:
        self._buffer.write(self.get_starttag_text() or "")

    def handle_startendtag(
        self,
        tag: str,
        attrs: list[tuple[str, Optional[str]]],
    ) -> None:
        self._buffer.write(self.get_starttag_text() or "")

    def handle_data(self, data: str) -> None:
        self._buffer.write(next(self._text))

    def handle_endtag(self, tag: str) -> None:
        self._buffer.write(f"</{tag}>")


def extract_text_from_html(html: str) -> list[str]:
    text = []
    extractor = HTMLTextExtractor()
    extractor.set_buffer(text)
    extractor.feed(html)
    return text


def inject_text_into_html(html: str, text: list[str]) -> str:
    buffer = io.StringIO()
    injector = HTMLTextInjector()
    injector.set_buffer(buffer)
    injector.set_text(text)
    injector.feed(html)
    return buffer.getvalue()


def extract_text_from_list(it: list) -> list[str]:
    def walk(it):
        for el in it:
            if isinstance(el, str):
                yield el
            else:
                yield from walk(el)

    return list(walk(it))


def inject_text_into_list(it: list, text: list[str]) -> list:
    def walk(it, text):
        for el in it:
            if isinstance(el, str):
                yield next(text)
            else:
                yield list(walk(el, text))

    return list(walk(it, iter(text)))



import unittest
from unittest import mock

from itemadapter.adapter import ItemAdapter
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.item import Field, Item
from scrapy.settings import Settings
from scrapy.spiders import Spider

from scrapy_translate.pipeline import (
    TranslatedFieldMeta,
    TranslatedString,
    TranslatePipeline,
)
from scrapy_translate.providers import (
    CacheProvider,
    IdentityTranslationProdiver,
    NullCacheProvider,
    TranslationProvider,
)


class TestTranslatePipeline(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.cache_mock = mock.create_autospec(CacheProvider, instance=True)
        self.translation_mock = mock.create_autospec(TranslationProvider, instance=True)

    def test_from_crawler(self):
        crawler_mock = mock.create_autospec(Crawler, instance=True)
        translation_provider = "scrapy_translate.providers.IdentityTranslationProdiver"
        type(crawler_mock).settings = mock.PropertyMock(
            return_value=Settings({"TRANSLATION_PROVIDER": translation_provider})
        )

        instance = TranslatePipeline.from_crawler(crawler_mock)

        self.assertIsInstance(instance, TranslatePipeline)
        self.assertIsInstance(instance._cache_provider, NullCacheProvider)
        self.assertIsInstance(
            instance._translation_provider,
            IdentityTranslationProdiver,
        )

    def test_from_crawler__translation_disabled__raise_not_configured(self):
        crawler_mock = mock.create_autospec(Crawler, instance=True)
        type(crawler_mock).settings = mock.PropertyMock(
            return_value=Settings({"TRANSLATION_DISABLED": True})
        )

        with self.assertRaises(NotConfigured):
            TranslatePipeline.from_crawler(crawler_mock)

    def test_from_crawler__provider_not_set__raise_not_configured(self):
        crawler_mock = mock.create_autospec(Crawler, instance=True)
        type(crawler_mock).settings = mock.PropertyMock()

        with self.assertRaises(NotConfigured):
            TranslatePipeline.from_crawler(crawler_mock)

    async def test_process_item(self):
        class TestItem(Item):
            field1 = Field(translate=True)
            field2 = Field(translate=True, translate_cache=True)
            field3 = Field(translate=True, translate_html=True, translate_cache=True)

        spider_mock = mock.create_autospec(Spider, instance=True)
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )
        self.cache_mock.get.return_value = {
            "dolor": "rolod",
            "consectetur": "rutetcesnoc",
        }
        self.translation_mock.translate.return_value = {
            "lorem": "merol",
            "ipsum": "muspi",
        }

        item = TestItem(
            field1="lorem",
            field2=["ipsum", "dolor"],
            field3="<p>sit amet, <strong>consectetur</strong></p>",
        )
        expected = TestItem(
            field1="merol",
            field2=["muspi", "rolod"],
            field3="<p>sit amet, <strong>rutetcesnoc</strong></p>",
        )

        actual = await pipeline.process_item(item, spider_mock)

        self.assertEqual(actual, expected)

    async def test_process_item__skip_not_items(self):
        spider_mock = mock.create_autospec(Spider, instance=True)
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        actual = await pipeline.process_item(..., spider_mock)

        self.assertIs(actual, ...)

    def test_translated_fields_meta(self):
        class TestItem(Item):
            field1 = Field()
            field2 = Field(translate=False)
            field3 = Field(translate=True)
            field4 = Field(translate=True, translate_html=True)
            field5 = Field(translate=True, translate_cache=True)
            field6 = Field(
                translate=False,
                translate_html=True,
                translate_cache=True,
            )

        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        item = TestItem()
        expected = [
            ("field3", TranslatedFieldMeta(html=False, cache=False)),
            ("field4", TranslatedFieldMeta(html=True, cache=False)),
            ("field5", TranslatedFieldMeta(html=False, cache=True)),
        ]

        actual = list(pipeline._translated_fields_meta(ItemAdapter(item)))

        self.assertEqual(actual, expected)

    def test_get_translated_strings(self):
        checks = (
            (Field(), "foobar", []),
            (Field(translate=True), "", []),
            (Field(translate=True), "foobar", ["foobar"]),
            (Field(translate=True), ["foo", "bar"], ["foo", "bar"]),
            (
                Field(translate=True, translate_html=True),
                "<p>foo<i>bar</i></p>",
                ["foo", "bar"],
            ),
        )
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        for field, value, expected in checks:
            with self.subTest(field=field, value=value):
                TestItem = type("TestItem", (Item,), {"field": field})
                item = TestItem(field=value)

                actual = pipeline._get_translated_strings(ItemAdapter(item))

                self.assertEqual(actual, expected)

    def test_get_translated_strings__set_string_meta(self):
        class TestItem(Item):
            field1 = Field(translate=True)
            field2 = Field(translate=True, translate_cache=True)

        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        item = TestItem(field1="foobar", field2=["foo", "bar"])

        strings = pipeline._get_translated_strings(ItemAdapter(item))

        self.assertEqual(strings, ["foobar", "foo", "bar"])
        cache_meta = [string.cache for string in strings]
        self.assertEqual(cache_meta, [False, True, True])

    def test_set_translated_strings(self):
        checks = (
            (Field(), "foo", "foo"),
            (Field(translate=True), "foo", "bar"),
            (Field(translate=True), ["foo", "baz"], ["bar", "qux"]),
            (
                Field(translate=True, translate_html=True),
                "<p>foo<i>baz</i></p>",
                "<p>bar<i>qux</i></p>",
            ),
        )

        translations = {"foo": "bar", "baz": "qux"}

        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        for field, value, expected in checks:
            with self.subTest(field=field, value=value):
                TestItem = type("TestItem", (Item,), {"field": field})
                item = TestItem(field=value)

                pipeline._set_translated_strings(
                    ItemAdapter(item),
                    translations,
                )

                self.assertEqual(item["field"], expected)

    def test_set_translated_strings__fallback_to_original(self):
        checks = (
            (Field(translate=True), "foo"),
            (Field(translate=True), ["foo", "baz"]),
            (Field(translate=True, translate_html=True), "<p>foo<i>baz</i></p>"),
        )

        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )

        for field, value in checks:
            with self.subTest(field=field, value=value):
                TestItem = type("TestItem", (Item,), {"field": field})
                item = TestItem(field=value)

                pipeline._set_translated_strings(ItemAdapter(item), {})

                self.assertEqual(item["field"], item["field"])

    async def test_translate(self):
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )
        self.cache_mock.get.return_value = {}
        self.translation_mock.translate.return_value = {"lorem": "merol"}

        strings = [TranslatedString("lorem")]
        expected = {"lorem": "merol"}

        actual = await pipeline._translate(strings)

        self.assertEqual(actual, expected)
        self.translation_mock.translate.assert_called_once_with({"lorem"})

    async def test_translate__get_from_cache(self):
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )
        self.cache_mock.get.return_value = {"lorem": "merol"}
        self.translation_mock.translate.return_value = {}

        strings = [TranslatedString("lorem", cache=True)]
        expected = {"lorem": "merol"}

        actual = await pipeline._translate(strings)

        self.assertEqual(actual, expected)
        self.cache_mock.get.assert_called_once_with({"lorem"})
        self.translation_mock.translate.assert_not_called()
        self.cache_mock.set.assert_not_called()

    async def test_translate__set_cache(self):
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )
        self.cache_mock.get.return_value = {}
        self.translation_mock.translate.return_value = {"lorem": "merol"}

        strings = [TranslatedString("lorem", cache=True)]
        expected = {"lorem": "merol"}

        actual = await pipeline._translate(strings)

        self.assertEqual(actual, expected)
        self.cache_mock.set.assert_called_once_with({"lorem": "merol"})

    async def test_translate__skip_cache(self):
        pipeline = TranslatePipeline(
            cache_provider=self.cache_mock,
            translation_provider=self.translation_mock,
        )
        self.cache_mock.get.return_value = {}
        self.translation_mock.translate.return_value = {"lorem": "merol"}

        strings = [TranslatedString("lorem")]
        expected = {"lorem": "merol"}

        actual = await pipeline._translate(strings)

        self.assertEqual(actual, expected)
        self.cache_mock.get.assert_not_called()
        self.translation_mock.translate.assert_called_once_with({"lorem"})
        self.cache_mock.set.assert_not_called()



from typing import Collection, Protocol


class CacheProvider(Protocol):
    async def get(self, strings: Collection[str], /) -> dict[str, str]: ...

    async def set(self, strings: dict[str, str], /) -> None: ...


class NullCacheProvider(CacheProvider):
    async def get(self, _: Collection[str]) -> dict[str, str]:
        return {}

    async def set(self, _: dict[str, str]) -> None:
        pass


class TranslationProvider(Protocol):
    async def translate(self, strings: Collection[str], /) -> dict[str, str]: ...


class IdentityTranslationProdiver(TranslationProvider):
    async def translate(self, strings: Collection[str]) -> dict[str, str]:
        return {string: string for string in strings}
